#undef TEST_WIN32

#include <sqVirtualMachine.h>
#include <surfaceplugin.h>

/* resides in MyFBPlugin.c */
extern struct VirtualMachine *interpreterProxy;

#define FB_WIDTH 640
#define FB_HEIGHT 480
#define FB_DEPTH 16

/* Entry points for the surface manager. Looked up on startup */
static fn_ioRegisterSurface registerSurface = 0;
static fn_ioUnregisterSurface unregisterSurface = 0;
static fn_ioFindSurface findSurface = 0;


/* the following four functions need to be implemented for surfaces */
static int fbGetSurfaceFormat(int handle, int* width, int* height, int* depth, int* isMSB);
static int fbLockSurface(int handle, int *pitch, int x, int y, int w, int h);
static int fbUnlockSurface(int handle, int x, int y, int w, int h);
static int fbShowSurface(int handle, int x, int y, int w, int h);

/* this structure is used to register surfaces */
static sqSurfaceDispatch fbDisplayDispatch = {
  1,
  0,
  (fn_getSurfaceFormat) fbGetSurfaceFormat,
  (fn_lockSurface) fbLockSurface,
  (fn_unlockSurface) fbUnlockSurface,
  (fn_showSurface) fbShowSurface
};

static int fbHandle = 0;
static int fbPointer = 0;

/* this function returns status information */
static int fbGetSurfaceFormat(int handle, int* width, int* height, int* depth, int* isMSB) {
  *width = FB_WIDTH;
  *height = FB_HEIGHT;
  *depth = FB_DEPTH;
  *isMSB = 0; /* <--- check this */
  return 1;
}

/* this function must aquire a hardware lock if needed */
static int fbLockSurface(int handle, int *pitch, int x, int y, int w, int h) {
  /* bytes per scanline need to be filled in */
  *pitch = FB_WIDTH * (FB_DEPTH / 8); /* <--- check this */
  return fbPointer;
}

/* this function must release a hardware lock if aquired */
static int fbUnlockSurface(int handle, int x, int y, int w, int h) {
  /* nothing to do */
  return 1;
}

/* this function must flush the bits on the screen if double buffered */
static int fbShowSurface(int handle, int x, int y, int w, int h) {
#ifdef TEST_WIN32
  drawBits();
#endif
  return 1;
}

/* these functions are called from the plugin */
int frameBufferAvailable(void) {
  return fbPointer != 0;
}
int frameBufferWidth(void) {
  return FB_WIDTH;
}
int frameBufferHeight(void) {
  return FB_HEIGHT;
}
int frameBufferDepth(void) {
  return FB_DEPTH;
}
int frameBufferHandle(void) {
  return fbHandle;
}
int frameBufferColorMasks(unsigned int *masks) {
#ifdef TEST_WIN32
  /* for testing swap red and green */
  masks[1] = (31 << 10); /* red mask */
  masks[0] = (31 << 5); /* green mask */
  masks[2] = (31 << 0); /* blue mask */
  masks[3] = 0; /* alpha mask */
  return 1;
#endif
  return 0; /* e.g., like in Squeak */
}

int frameBufferClose(void) {
  (*unregisterSurface)(fbHandle);
  free(fbPointer);
}

int frameBufferInit(void) {
  /* initialize the frame buffer */
  fbPointer = (int) malloc(FB_WIDTH * FB_HEIGHT * FB_DEPTH / 8);
  if(!fbPointer) return 0;

  /* now do the house-keeping */

  /* look up the entries from the VM */
  registerSurface = (fn_ioRegisterSurface) 
    interpreterProxy->ioLoadFunctionFrom("ioRegisterSurface","SurfacePlugin");
  unregisterSurface = (fn_ioUnregisterSurface)
    interpreterProxy->ioLoadFunctionFrom("ioUnregisterSurface","SurfacePlugin");
  findSurface = (fn_ioFindSurface)
    interpreterProxy->ioLoadFunctionFrom("ioFindSurface","SurfacePlugin");
  if(!registerSurface || !unregisterSurface || !findSurface) {
    printf("ERROR: Surface functions not found\n");
    /* this would be the place to abort() or exit() */
    return 0;
  }

  /* register the frame buffer */
  if((*registerSurface)(0 /*unused*/, &fbDisplayDispatch, &fbHandle) == 0) {
    printf("ERROR: FB registration failed\n");
    /* this would be the place to abort() or exit() */
    return 0;
  }
  /* okay, now we're all set up */
  return 1;
}


/* test stuff for win32.
   Note: What follows is so stupid that you _definitely_ don't want to do it that way. It's just for me to see if the plugin stuff is actually working. */
#ifdef TEST_WIN32
#include <windows.h>

static HWND myHwnd = 0;
static BITMAPINFO *bmi16;

int drawBits(void) {
  HDC dc;
  if(!myHwnd) myHwnd = GetFocus(); /* this might be the wrong window */
  if(!bmi16) {
    bmi16 = (BITMAPINFO*) calloc(1,sizeof(BITMAPINFO) + 4 * sizeof(DWORD));
    bmi16->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    bmi16->bmiHeader.biPlanes = 1;
    bmi16->bmiHeader.biBitCount = 16;
    bmi16->bmiHeader.biCompression = BI_BITFIELDS;
    bmi16->bmiHeader.biWidth = FB_WIDTH;
    bmi16->bmiHeader.biHeight = -FB_HEIGHT;
    bmi16->bmiHeader.biSizeImage = 0;
    ((DWORD*) bmi16->bmiColors)[0] = (31 << 10); /* red mask */
    ((DWORD*) bmi16->bmiColors)[1] = (31 << 5); /* green mask */
    ((DWORD*) bmi16->bmiColors)[2] = (31 << 0); /* blue mask */
  }
  dc = GetDC(myHwnd);
  SetDIBitsToDevice(dc, 0, 0, FB_WIDTH, FB_HEIGHT, 0, 0, 0, FB_HEIGHT, (void*) fbPointer, bmi16, DIB_RGB_COLORS);
  ReleaseDC(myHwnd,dc);
}
#endif
